#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from __future__ import annotations

from typing import Any, Mapping, MutableMapping

StreamSlice = Mapping[str, Any]
StreamState = MutableMapping[str, Any]
